/* Line clipping and drawing, hereby released into public domain. */
/* Originally written by Eric Auer in 1997. Have fun! Eric, 2003. */

#define XGFX 320
#define YGFX 200
#typedef double FP

/* A direct pointer to the MCGA screen buffer: */
byte far *mcga_r=PTR(0xa000,0);

/* You could ...=(byte *)farmalloc(XGFX * YGFX) bytes which you  */
/* farfree(...) later, as a buffer. Or use MCGA RAM directly:    */
byte far *act_pg = &mcga_r[0]; /* active page (to which we draw) */

/* You could use this as a pointer into some buffer, if you do   */
/* the copying act_pg to mcga_r by hand. If you use a video mode */
/* where the hardware allows several pages, this is not needed.  */
/* byte far *vis_pg; --- page to be displayed */

/* You find (an array of) word(s) with cursor coordinates in: */
word volatile far *textcursor=PTR(0x40,0x50);
/* You find a keyboard shift bitmap in: */
word volatile far *keyshifts=PTR(0x40,0x17);
byte volatile far *gfxmode=PTR(0x40,0x49);

#define MAX(a,b) (((a)>(b)) ? a : b)
#define tmpSWAP(a,b,tmp) tmp=a; a=b; b=tmp

/* Set a pixel at x,y to color c. */
#define MCGAACT(x,y,c) act_pg[ ((y) * XGFX) + x] = c

#define CLIP_ENABLE false
#define CLIP_NEEDED(x1,y1,x2,y2) (CLIP_ENABLE &&  \
 (((x1)<0) || ((x1)>=XGFX) || ((y1)<0) || ((y1)>=YGFX) || \
  ((x2)<0) || ((x2)>=XGFX) || ((y2)<0) || ((y2)>=YGFX)) )

/* Assumes XGFX and YGFX to be the screen resolution */
/* FP is the floating point type of your choice, double or float. */

/* Adjust the line from x1,y1 to x2,y2 so that it will end at the  */
/* border of your screen. Returns 0,0 to 0,0 if all is off screen. */
/* */
void clip_line(int far *x1, int far *y1, int far *x2, int far *y2) {
  FP slope; /*y(x)=y1+(slope*x) x(y)=...*/

  /* correct offscale X coordinates, if not vertical. */
  if (*x1!=*x2) { /* if not vertical... */
    slope = (0.0 + *y2 - *y1) / (*x2 - *x1);
   /* the 0.0 is needed to select floating point mode! */
    if (*x1<0) { /* too far left? */
      *y1 += round( slope * (0.0 - *x1) );
      *x1 = 0;
    };
    if (*x2<0) { /* too far left? */
      *y2 = *y1 + round( slope * (0.0 - *x1) );
      *x2 = 0;
    };
    if (*x1>=XGFX) { /* too far right? */
      *y1 += round( slope * (-1.0 + XGFX - *x1) );
      *x1 = XGFX - 1;
    };
    if (*x2>=XGFX) { /* too far right? */
      *y2 = *y1 + round( slope * (-1.0 + XGFX - *x1) );
      *x2 = XGFX - 1;
    };
  } else { /* vertical. Check if visible. */
    if ((*x1<0) || (*x1>=XGFX) || (*x2<0) || (*x2>=XGFX)) {
      *x1 = 0;
      *x2 = 0;
      *y1 = 0;
      *y2 = 0;
    };

  /* correct offscale Y coordinates, if not horizontal. */
  if (*y1!=*y2) { /* if not horizontal... */
    slope = (0.0 + *x2 - *x1) / (*y2 - *y1);
    /* the 0.0 is needed to select floating point mode! */
    if (*y1<0) { /* above screen? */
      *x1 += round( slope*(0.0 - *y1) );
      *y1 = 0;
    };
    if (*y2<0) { /* above screen? */
      *x2 = *x1 + round( slope * (0.0 - *y1) );
      *y2 = 0;
    };
    if (*y1>=YGFX) { /* below screen? */
      *x1 += round( slope * (-1.0 + YGFX - *y1) );
      *y1 = YGFX - 1;
    };
    if (*y2>=YGFX) {
      *x2 = *x1 + round( slope * (-1.0+YGFX- *y1) );
      *y2 = YGFX - 1;
    };
  } else  { /* Horizontal. Check if visible. */
    if ((*y1<0) || (*y1>=YGFX) || (*y2<0) || (*y2>=YGFX)) {
      *x1=0; *x2=0; *y1=0; *y2=0; /*ganze Linie offscale*/
    };
  };
}; /* clip_line */



/* A standard Bresenham algorithm line drawing implementation */
/* Bresenham represents the slope as a fraction: You count in */
/* 1/Nths of pixels in one variable, and overflow values of N */
/* or bigger by into the actual pixel coordinate values...    */
/* */
void mcga_line(word ix1, word iy1, word ix2, word iy2, word c) {
  register int _cx, _cy;
  int x1, y1, x2, y2, dx, dy, d2, sgn, c2;

  ok=false; x1=ix1; x2=ix2; y1=iy1; y2=iy2; c2=c;

  if (CLIP_NEEDED(x1, y1, x2, y2))
    clip_line(&x1, &y1, &x2, &y2);

  if ((x1==x2) && (y1==y2)) return;

  if (y1==y2) {
    if (x1>x2)
      tmpSWAP(x1, x2, _cx);
    for (_cx=x1; _cx<=x2; _cx++)
      MCGAACT(_cx, y1, c);
    return;
  };
  if (x1==x2) {
    if (y1>y2)
      tmpSWAP(y1, y2, _cy);
    for (_cy=y1; _cy<=y2; _cy++)
      MCGAACT(x1,_cy)=c;
    return;
  };

  /* That was the obvious part. Next is the Bresenham part. */
  dx = x2 - x1;
  dy = y2 - y1;
  d2=0;
  if (abs(dx)>abs(dy)) { /* loop x */
    /* We iterate through the X values, line is relatively flat. */
    sgn = (dy<0) ? -1 : 1;
    dy = abs(dy);
    if (dx>0) { /* left to right */
      /* Change y by (dy/dx) each iteration: Done by changing d2 */
      /* by dy and changing y by one every time when d2 > dx...  */
      /* then,  correct d2 by dx again, and so on. Could add yet */
      /* another IF THEN ELSE condition here and remove sgn...   */
      _cy = y1;
      for (_cx=x1; _cx<=x2; _cx++) {
        MCGAACT(_cx,_cy, c2);
        d2 += dy;
        if (d2>=dx) {
          d2 -= dx; /* correct sub pixel count */
          _cy += sgn; /* 1 pixel step */
        };
      }; /* for */
    } else { /* dx<0, right to left */
      dx = -dx;
      sgn = -sgn;
      _cy = y2;
      for (_cx=x2; _cx<=x1; _cx++) {
        MCGAACT(_cx,_cy, c2);
        d2 += dy;
        if (d2>dx) {
          d2 -= dx; /* correct sub pixel count */
          _cy += sgn; /* 1 pixel step */
        };
      }; /* for */
    };
  } else { /* loop y */
    /* We iterate through the Y values, line is relatively steep. */
    /* Everything works more or less as above otherwise. */
    sgn = (dx<0) ? -1 : 1;
    dx = abs(dx);
    if (dy>0) { /* top to bottom */
      _cx=x1;
      for (_cy=y1; _cy<=y2; _cy++) {
        MCGAACT(_cx,_cy, c2);
        d2 += dx;
        if (d2>=dy) {
          d2 -= dy;
          _cx += sgn;
        };
      }; /* for */
    } else { /* bottom to top */
      dy = -dy;
      sgn = -sgn;
      _cx = x2;
      for (_cy=y2; _cy<=y1; _cy++) {
        MCGAACT(_cx,_cy, c2);
        d2 += dx;
        if (d2>dy) {
          d2 -= dy;
          _cx += sgn;
        };
      }; /* for */
    };
  }; /* loop y */
}; /* mcga_line */

